/*
 * Copyright (c) 2009, 2015, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.sun.btrace.runtime;

import com.sun.btrace.instr.RandomIntProvider;
import java.lang.reflect.Field;
import org.junit.BeforeClass;
import support.InstrumentorTestBase;
import org.junit.Test;

/**
 *
 * @author Jaroslav Bachorik
 */
public class InstrumentorTest extends InstrumentorTestBase {
    @BeforeClass
    public static void classSetup() throws Exception {
        try {
            Field f = RandomIntProvider.class.getDeclaredField("useBtraceEnter");
            f.setAccessible(true);
            f.setBoolean(null, false);
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    @Test
    public void matchDerivedClass() throws Exception {
        loadTargetClass("DerivedClass");
        transform("onmethod/MatchDerived");

        checkTransformation("ALOAD 0\nALOAD 1\nALOAD 2\n"
                + "INVOKESTATIC resources/DerivedClass.$btrace$traces$onmethod$MatchDerived$args (Lresources/AbstractClass;Ljava/lang/String;Ljava/util/Map;)V");
    }

    @Test
    public void methodEntryCheckcastBefore() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/CheckcastBefore");

        checkTransformation("DUP\nASTORE 2\nALOAD 0\nLDC \"resources.OnMethodTest\"\nALOAD 2\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$CheckcastBefore$args (Ljava/lang/Object;Ljava/lang/String;Ljava/util/HashMap;)V");
    }

    @Test
    public void methodEntryCheckcastAfter() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/CheckcastAfter");

        checkTransformation("DUP\nALOAD 0\nLDC \"casts\"\nALOAD 2\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$CheckcastAfter$args (Ljava/lang/Object;Ljava/lang/String;Ljava/util/HashMap;)V\n");
    }

    @Test
    public void methodEntryInstanceofBefore() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/InstanceofBefore");

        checkTransformation("DUP\nASTORE 3\nALOAD 0\nLDC \"resources.OnMethodTest\"\nALOAD 3\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$InstanceofBefore$args (Ljava/lang/Object;Ljava/lang/String;Ljava/util/HashMap;)V");
    }

    @Test
    public void methodEntryInstanceofAfter() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/InstanceofAfter");

        checkTransformation("DUP\nASTORE 3\nALOAD 0\nLDC \"casts\"\nALOAD 3\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$InstanceofAfter$args (Ljava/lang/Object;Ljava/lang/String;Ljava/util/HashMap;)V\n");
    }

    @Test
    public void methodEntryCatch() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/Catch");

        checkTransformation("DUP\nALOAD 0\nALOAD 1\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$Catch$args (Ljava/lang/Object;Ljava/io/IOException;)V\n"
                + "ASTORE 2");
    }

    @Test
    public void methodEntryThrow() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/Throw");

        checkTransformation("DUP\nASTORE 1\nALOAD 0\nLDC \"resources.OnMethodTest\"\nLDC \"exception\"\nALOAD 1\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$Throw$args (Ljava/lang/Object;Ljava/lang/String;Ljava/lang/String;Ljava/lang/Throwable;)V");
    }

    @Test
    public void methodEntryError() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/Error");

        checkTransformation("TRYCATCHBLOCK L0 L1 L1 java/lang/Throwable\n"
                + "DUP\nASTORE 1\nALOAD 0\nLDC \"uncaught\"\nALOAD 1\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$Error$args (Ljava/lang/Object;Ljava/lang/String;Ljava/lang/Throwable;)V");
    }

    @Test
    public void methodEntryErrorDuration() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/ErrorDuration");

        checkTransformation(
            "TRYCATCHBLOCK L0 L1 L1 java/lang/Throwable\n" +
            "LDC 0\n" +
            "LSTORE 1\n" +
            "INVOKESTATIC java/lang/System.nanoTime ()J\n" +
            "LSTORE 3\n" +
            "INVOKESTATIC java/lang/System.nanoTime ()J\n" +
            "LLOAD 3\n" +
            "LSUB\n" +
            "LSTORE 1\n" +
            "DUP\n" +
            "ASTORE 5\n" +
            "ALOAD 0\n" +
            "LDC \"uncaught\"\n" +
            "LLOAD 1\n" +
            "ALOAD 5\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$ErrorDuration$args (Ljava/lang/Object;Ljava/lang/String;JLjava/lang/Throwable;)V\n" +
            "ATHROW\n" +
            "MAXSTACK = 6\n" +
            "MAXLOCALS = 6");
    }

    @Test
    public void methodEntryLine() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/Line");

        checkTransformation("LDC \"field\"\nLDC 84\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$Line$args (Ljava/lang/Object;Ljava/lang/String;I)V\n"
                + "ALOAD 0");
    }

    @Test
    public void methodEntryNewBefore() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/NewBefore");

        checkTransformation("ALOAD 0\nLDC \"java.util.HashMap\"\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$NewBefore$args (Ljava/lang/Object;Ljava/lang/String;)V");
    }

    @Test
    public void methodEntryNewAfter() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/NewAfter");

        checkTransformation("ASTORE 1\nALOAD 0\nALOAD 1\nLDC \"java.util.HashMap\"\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$NewAfter$args (Ljava/lang/Object;Ljava/util/Map;Ljava/lang/String;)V\n"
                + "DUP");
    }

    @Test
    public void methodEntrySyncEntry() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/SyncEntry");

        checkTransformation("TRYCATCHBLOCK L4 L5 L5 java/lang/Throwable\n"
                + "DUP\nASTORE 2\nALOAD 0\nLDC \"sync\"\nALOAD 2\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$SyncEntry$args (Ljava/lang/Object;Ljava/lang/String;Ljava/lang/Object;)V");
    }

    @Test
    public void methodEntrySyncExit() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/SyncExit");

        checkTransformation("TRYCATCHBLOCK L4 L5 L5 java/lang/Throwable\n"
                + "L6\nLINENUMBER 110 L6\n"
                + "DUP\nASTORE 2\nALOAD 0\nLDC \"resources/OnMethodTest\"\nALOAD 2\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$SyncExit$args (Ljava/lang/Object;Ljava/lang/String;Ljava/lang/Object;)V\n");
    }

    @Test
    public void methodEntryNewArrayIntBefore() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/NewArrayIntBefore");

        checkTransformation("ALOAD 0\nLDC \"int\"\nLDC 1\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$NewArrayIntBefore$args (Ljava/lang/Object;Ljava/lang/String;I)V");
    }

    @Test
    public void methodEntryNewArrayStringBefore() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/NewArrayStringBefore");

        checkTransformation("ALOAD 0\nLDC \"java.lang.String\"\nLDC 1\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$NewArrayStringBefore$args (Ljava/lang/Object;Ljava/lang/String;I)V");
    }

    @Test
    public void methodEntryNewArrayIntAfter() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/NewArrayIntAfter");

        checkTransformation("DUP\nALOAD 0\nALOAD 1\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$NewArrayIntAfter$args (Ljava/lang/Object;[I)V");
    }

    @Test
    public void methodEntryNewArrayStringAfter() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/NewArrayStringAfter");

        checkTransformation("DUP\nALOAD 0\nALOAD 3\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$NewArrayStringAfter$args (Ljava/lang/Object;[Ljava/lang/String;)V");
    }

    @Test
    public void methodEntryArrayGetBefore() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/ArrayGetBefore");

        checkTransformation("DUP2\nISTORE 3\nASTORE 4\nALOAD 0\nALOAD 4\nILOAD 3\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$ArrayGetBefore$args (Ljava/lang/Object;[II)V");
    }

    @Test
    public void methodEntryArrayGetAfter() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/ArrayGetAfter");

        checkTransformation("DUP2\nISTORE 3\nASTORE 4\nDUP\nISTORE 5\nALOAD 0\nILOAD 5\nALOAD 4\nILOAD 3\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$ArrayGetAfter$args (Ljava/lang/Object;I[II)V");
    }

    @Test
    public void methodEntryArraySetBefore() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/ArraySetBefore");

        checkTransformation("ISTORE 4\nDUP2\nISTORE 5\nASTORE 6\nILOAD 4\nALOAD 0\nALOAD 6\nILOAD 5\nILOAD 4\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$ArraySetBefore$args (Ljava/lang/Object;[III)V");
    }

    @Test
    public void methodEntryArraySetAfter() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/ArraySetAfter");

        checkTransformation("ISTORE 4\nDUP2\nISTORE 5\nASTORE 6\nILOAD 4\nALOAD 0\nALOAD 6\nILOAD 5\nILOAD 4\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$ArraySetAfter$args (Ljava/lang/Object;[III)V");
    }

    @Test
    public void methodEntryFieldGetBefore() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/FieldGetBefore");

        checkTransformation("DUP\nASTORE 1\nALOAD 0\nALOAD 1\nLDC \"field\"\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$FieldGetBefore$args (Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/String;)V");
    }

    @Test
    public void methodEntryFieldGetAfter() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/FieldGetAfter");

        checkTransformation("DUP\nASTORE 1\nDUP\nISTORE 2\nALOAD 0\nALOAD 1\nLDC \"field\"\nILOAD 2\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$FieldGetAfter$args (Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/String;I)V");
    }

    @Test
    public void methodEntryFieldSetBefore() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/FieldSetBefore");

        checkTransformation("ISTORE 1\nDUP\nASTORE 2\nILOAD 1\nALOAD 0\nALOAD 2\nLDC \"field\"\nILOAD 1\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$FieldSetBefore$args (Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/String;I)V");
    }

    @Test
    public void methodEntryFieldSetAfter() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/FieldSetAfter");

        checkTransformation("ISTORE 1\nDUP\nASTORE 2\nILOAD 1\nALOAD 0\nALOAD 2\nLDC \"field\"\nILOAD 1\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$FieldSetAfter$args (Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/String;I)V");
    }

    @Test
    public void methodEntryArgsNoSelf() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/ArgsNoSelf");

        checkTransformation("ALOAD 1\nLLOAD 2\nALOAD 4\nALOAD 5\nINVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$ArgsNoSelf$argsNoSelf (Ljava/lang/String;J[Ljava/lang/String;[I)V");
    }

    @Test
    public void methodEntryNoArgs() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/NoArgs");

        checkTransformation("ALOAD 0\nINVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$NoArgs$argsEmpty (Ljava/lang/Object;)V");
    }

    @Test
    public void methodEntryArgs() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/Args");
        checkTransformation("ALOAD 0\nALOAD 1\nLLOAD 2\nALOAD 4\nALOAD 5\nINVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$Args$args (Ljava/lang/Object;Ljava/lang/String;J[Ljava/lang/String;[I)V");
    }

    @Test
    public void methodEntryArgsSampledNoSampling() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/ArgsSampledNoSampling");
        checkTransformation("LDC 5\n" +
            "INVOKESTATIC com/sun/btrace/instr/MethodTracker.hit (I)Z\n" +
            "ISTORE 6\n" +
            "ILOAD 6\n" +
            "IFEQ L0\n" +
            "ALOAD 0\n" +
            "ALOAD 1\n" +
            "LLOAD 2\n" +
            "ALOAD 4\n" +
            "ALOAD 5\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$ArgsSampledNoSampling$argsSampled (Ljava/lang/Object;Ljava/lang/String;J[Ljava/lang/String;[I)V\n" +
            "ALOAD 0\n" +
            "ALOAD 1\n" +
            "LLOAD 2\n" +
            "ALOAD 4\n" +
            "ALOAD 5\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$ArgsSampledNoSampling$argsNoSampling (Ljava/lang/Object;Ljava/lang/String;J[Ljava/lang/String;[I)V\n" +
            "L1\n" +
            "LINENUMBER 44 L1\n" +
            "L2");
    }

    @Test
    public void methodEntryArgsSampled() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/ArgsSampled");
        checkTransformation("LDC 5\n" +
            "INVOKESTATIC com/sun/btrace/instr/MethodTracker.hit (I)Z\n" +
            "ISTORE 6\n" +
            "ILOAD 6\n" +
            "IFEQ L0\n" +
            "ALOAD 0\n" +
            "ALOAD 1\n" +
            "LLOAD 2\n" +
            "ALOAD 4\n" +
            "ALOAD 5\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$ArgsSampled$args (Ljava/lang/Object;Ljava/lang/String;J[Ljava/lang/String;[I)V\n" +
            "MAXSTACK = 6\n" +
            "MAXLOCALS = 7"
        );
    }

    @Test
    public void methodEntryArgs2Sampled() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/Args2Sampled");
        checkTransformation("LDC 5\n" +
            "INVOKESTATIC com/sun/btrace/instr/MethodTracker.hit (I)Z\n" +
            "ISTORE 6\n" +
            "ILOAD 6\n" +
            "IFEQ L0\n" +
            "ALOAD 0\n" +
            "ALOAD 1\n" +
            "LLOAD 2\n" +
            "ALOAD 4\n" +
            "ALOAD 5\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$Args2Sampled$args2 (Ljava/lang/Object;Ljava/lang/String;J[Ljava/lang/String;[I)V\n" +
            "ILOAD 6\n" +
            "IFEQ L1\n" +
            "ALOAD 0\n" +
            "ALOAD 1\n" +
            "LLOAD 2\n" +
            "ALOAD 4\n" +
            "ALOAD 5\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$Args2Sampled$args (Ljava/lang/Object;Ljava/lang/String;J[Ljava/lang/String;[I)V\n" +
            "L1\n" +
            "LINENUMBER 44 L1\n" +
            "L2"
        );
    }

    @Test
    public void methodEntryArgsSampledAdaptive() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/ArgsSampledAdaptive");
        checkTransformation("LDC 5\n" +
            "INVOKESTATIC com/sun/btrace/instr/MethodTracker.hitAdaptive (I)Z\n" +
            "ISTORE 6\n" +
            "ILOAD 6\n" +
            "IFEQ L0\n" +
            "ALOAD 0\n" +
            "ALOAD 1\n" +
            "LLOAD 2\n" +
            "ALOAD 4\n" +
            "ALOAD 5\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$ArgsSampledAdaptive$args (Ljava/lang/Object;Ljava/lang/String;J[Ljava/lang/String;[I)V\n" +
            "ILOAD 6\n" +
            "IFEQ L1\n" +
            "LDC 5\n" +
            "INVOKESTATIC com/sun/btrace/instr/MethodTracker.updateEndTs (I)V\n" +
            "L1\n" +
            "L2"
        );
    }

    @Test
    public void methodEntryArgsReturn() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/ArgsReturn");
        checkTransformation("DUP2\nLSTORE 6\nALOAD 0\nLLOAD 6\nALOAD 1\nLLOAD 2\nALOAD 4\nALOAD 5\nINVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$ArgsReturn$args (Ljava/lang/Object;JLjava/lang/String;J[Ljava/lang/String;[I)V");
    }

    @Test
    public void methodEntryArgsReturnAugmented() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/ArgsReturnAugmented", true);
        checkTransformation("LSTORE 6\n" +
            "ALOAD 0\n" +
            "ALOAD 1\n" +
            "LLOAD 2\n" +
            "ALOAD 4\n" +
            "ALOAD 5\n" +
            "LLOAD 6\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$ArgsReturnAugmented$args (Ljava/lang/Object;Ljava/lang/String;J[Ljava/lang/String;[IJ)J\n" +
            "MAXSTACK = 8\n" +
            "MAXLOCALS = 8\n" +
            "\n" +
            "// access flags 0xA\n" +
            "private static $btrace$traces$onmethod$ArgsReturnAugmented$args(Ljava/lang/Object;Ljava/lang/String;J[Ljava/lang/String;[IJ)J\n" +
            "@Lcom/sun/btrace/annotations/OnMethod;(clazz=\"/.*\\\\.OnMethodTest/\", method=\"args\", location=@Lcom/sun/btrace/annotations/Location;(value=Lcom/sun/btrace/annotations/Kind;.RETURN))\n" +
            "@Lcom/sun/btrace/annotations/Self;() // parameter 0\n" +
            "@Lcom/sun/btrace/annotations/Return;() // parameter 5\n" +
            "TRYCATCHBLOCK L0 L1 L1 java/lang/Throwable\n" +
            "GETSTATIC traces/onmethod/ArgsReturnAugmented.runtime : Lcom/sun/btrace/BTraceRuntime;\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.enter (Lcom/sun/btrace/BTraceRuntime;)Z\n" +
            "IFNE L0\n" +
            "LLOAD 6\n" +
            "LRETURN\n" +
            "L0\n" +
            "LDC \"args\"\n" +
            "INVOKESTATIC com/sun/btrace/BTraceUtils.println (Ljava/lang/Object;)V\n" +
            "LCONST_1\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "LRETURN\n" +
            "L1\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.handleException (Ljava/lang/Throwable;)V\n" +
            "LLOAD 6\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "LRETURN\n" +
            "MAXSTACK = 2\n" +
            "MAXLOCALS = 8"
        );
    }

    @Test
    public void methodEntryArgsReturnAugmented1() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/ArgsReturnAugmented1", true);
        checkTransformation("LSTORE 6\n" +
            "ALOAD 0\n" +
            "LLOAD 6\n" +
            "ALOAD 1\n" +
            "LLOAD 2\n" +
            "ALOAD 4\n" +
            "ALOAD 5\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$ArgsReturnAugmented1$args (Ljava/lang/Object;JLjava/lang/String;J[Ljava/lang/String;[I)J\n" +
            "MAXSTACK = 8\n" +
            "MAXLOCALS = 8\n" +
            "\n" +
            "// access flags 0xA\n" +
            "private static $btrace$traces$onmethod$ArgsReturnAugmented1$args(Ljava/lang/Object;JLjava/lang/String;J[Ljava/lang/String;[I)J\n" +
            "@Lcom/sun/btrace/annotations/OnMethod;(clazz=\"/.*\\\\.OnMethodTest/\", method=\"args\", location=@Lcom/sun/btrace/annotations/Location;(value=Lcom/sun/btrace/annotations/Kind;.RETURN))\n" +
            "@Lcom/sun/btrace/annotations/Self;() // parameter 0\n" +
            "@Lcom/sun/btrace/annotations/Return;() // parameter 1\n" +
            "TRYCATCHBLOCK L0 L1 L1 java/lang/Throwable\n" +
            "GETSTATIC traces/onmethod/ArgsReturnAugmented1.runtime : Lcom/sun/btrace/BTraceRuntime;\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.enter (Lcom/sun/btrace/BTraceRuntime;)Z\n" +
            "IFNE L0\n" +
            "LLOAD 1\n" +
            "LRETURN\n" +
            "L0\n" +
            "LDC \"args\"\n" +
            "INVOKESTATIC com/sun/btrace/BTraceUtils.println (Ljava/lang/Object;)V\n" +
            "LLOAD 1\n" +
            "LCONST_1\n" +
            "LADD\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "LRETURN\n" +
            "L1\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.handleException (Ljava/lang/Throwable;)V\n" +
            "LLOAD 1\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "LRETURN\n" +
            "MAXSTACK = 4\n" +
            "MAXLOCALS = 8"
        );
    }

    @Test
    public void methodEntryArgsReturnSampled() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/ArgsReturnSampled");
        checkTransformation("LDC 5\n" +
            "INVOKESTATIC com/sun/btrace/instr/MethodTracker.hit (I)Z\n" +
            "ISTORE 6\n" +
            "ILOAD 6\n" +
            "IFEQ L1\n" +
            "DUP2\n" +
            "LSTORE 7\n" +
            "ALOAD 0\n" +
            "LLOAD 7\n" +
            "ALOAD 1\n" +
            "LLOAD 2\n" +
            "ALOAD 4\n" +
            "ALOAD 5\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$ArgsReturnSampled$args (Ljava/lang/Object;JLjava/lang/String;J[Ljava/lang/String;[I)V\n" +
            "L1\n" +
            "L2"
        );
    }

    @Test
    public void methodEntryArgsDuration() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/ArgsDuration");
        checkTransformation("LDC 0\nLSTORE 6\n"
                + "INVOKESTATIC java/lang/System.nanoTime ()J\nLSTORE 8\n"
                + "INVOKESTATIC java/lang/System.nanoTime ()J\nLLOAD 8\n"
                + "LSUB\nLSTORE 6\nDUP2\nLSTORE 10\nALOAD 0\nLLOAD 10\nLLOAD 6\n"
                + "ALOAD 1\nLLOAD 2\nALOAD 4\nALOAD 5\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$ArgsDuration$args (Ljava/lang/Object;JJLjava/lang/String;J[Ljava/lang/String;[I)V\n"
                + "MAXSTACK");
    }

    @Test
    public void methodEntryArgsDurationMultiReturn() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/ArgsDurationMultiReturn");
        checkTransformation(
            "LDC 0\n" +
            "LSTORE 6\n" +
            "INVOKESTATIC java/lang/System.nanoTime ()J\n" +
            "LSTORE 8\n" +
            "INVOKESTATIC java/lang/System.nanoTime ()J\n" +
            "LLOAD 8\n" +
            "LSUB\n" +
            "LSTORE 6\n" +
            "DUP2\n" +
            "LSTORE 10\n" +
            "ALOAD 0\n" +
            "LLOAD 10\n" +
            "LLOAD 6\n" +
            "ALOAD 1\n" +
            "LLOAD 2\n" +
            "ALOAD 4\n" +
            "ALOAD 5\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$ArgsDurationMultiReturn$args (Ljava/lang/Object;JJLjava/lang/String;J[Ljava/lang/String;[I)V\n" +
            "INVOKESTATIC java/lang/System.nanoTime ()J\n" +
            "LLOAD 8\n" +
            "LSUB\n" +
            "LSTORE 6\n" +
            "DUP2\n" +
            "LSTORE 12\n" +
            "ALOAD 0\n" +
            "LLOAD 12\n" +
            "LLOAD 6\n" +
            "ALOAD 1\n" +
            "LLOAD 2\n" +
            "ALOAD 4\n" +
            "ALOAD 5\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$ArgsDurationMultiReturn$args (Ljava/lang/Object;JJLjava/lang/String;J[Ljava/lang/String;[I)V\n" +
            "INVOKESTATIC java/lang/System.nanoTime ()J\n" +
            "LLOAD 8\n" +
            "LSUB\n" +
            "LSTORE 6\n" +
            "DUP2\n" +
            "LSTORE 14\n" +
            "ALOAD 0\n" +
            "LLOAD 14\n" +
            "LLOAD 6\n" +
            "ALOAD 1\n" +
            "LLOAD 2\n" +
            "ALOAD 4\n" +
            "ALOAD 5\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$ArgsDurationMultiReturn$args (Ljava/lang/Object;JJLjava/lang/String;J[Ljava/lang/String;[I)V\n" +
            "MAXSTACK = 12\n" +
            "MAXLOCALS = 16");
    }

    @Test
    public void methodEntryArgsDurationSampled() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/ArgsDurationSampled");
        checkTransformation(
            "LDC 0\nLSTORE 6\nLDC 5\nINVOKESTATIC com/sun/btrace/instr/MethodTracker.hitTimed (I)J\n" +
            "DUP2\nLSTORE 8\nL2I\nISTORE 10\nILOAD 10\nIFEQ L1\n" +
            "LDC 5\nINVOKESTATIC com/sun/btrace/instr/MethodTracker.getEndTs (I)J\n" +
            "LLOAD 8\nLSUB\nLSTORE 6\nDUP2\nLSTORE 11\n" +
            "ALOAD 0\nLLOAD 11\nLLOAD 6\nALOAD 1\nLLOAD 2\nALOAD 4\nALOAD 5\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$ArgsDurationSampled$args (Ljava/lang/Object;JJLjava/lang/String;J[Ljava/lang/String;[I)V\n" +
            "L1\nL2\n"
        );
    }

    @Test
    public void methodEntryArgsDurationBoxed() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/ArgsDurationBoxed");
        checkTransformation("");
    }

    @Test
    public void methodEntryArgsDurationConstructor() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/ArgsDurationConstructor");
        checkTransformation("LDC 0\nLSTORE 2\n"
                + "INVOKESTATIC java/lang/System.nanoTime ()J\nLSTORE 4\n"
                + "INVOKESTATIC java/lang/System.nanoTime ()J\n"
                + "LLOAD 4\nLSUB\nLSTORE 2\n"
                + "ALOAD 0\nLLOAD 2\nALOAD 1\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$ArgsDurationConstructor$args (Ljava/lang/Object;JLjava/lang/String;)V\n"
                + "MAXSTACK");
    }

    @Test
    // check for multiple timestamps
    public void methodEntryArgsDuration2() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/ArgsDuration2");
        checkTransformation("LDC 0\nLSTORE 6\n"
                + "INVOKESTATIC java/lang/System.nanoTime ()J\nLSTORE 8\n"
                + "INVOKESTATIC java/lang/System.nanoTime ()J\n"
                + "LLOAD 8\nLSUB\nLSTORE 6\n"
                + "DUP2\nLSTORE 10\nALOAD 0\nLLOAD 10\nLLOAD 6\n"
                + "ALOAD 1\nLLOAD 2\nALOAD 4\nALOAD 5\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$ArgsDuration2$args2 (Ljava/lang/Object;JJLjava/lang/String;J[Ljava/lang/String;[I)V\n"
                + "DUP2\nLSTORE 12\nALOAD 0\nLLOAD 12\n"
                + "LLOAD 6\nALOAD 1\nLLOAD 2\nALOAD 4\nALOAD 5\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$ArgsDuration2$args (Ljava/lang/Object;JJLjava/lang/String;J[Ljava/lang/String;[I)V\n"
                + "MAXSTACK = 12\nMAXLOCALS = 14");
    }

    @Test
    // check for multiple timestamps
    public void methodEntryArgsDuration2Sampled() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/ArgsDuration2Sampled");
        checkTransformation("LDC 0\nLSTORE 6\nLDC 5\n"
                + "INVOKESTATIC com/sun/btrace/instr/MethodTracker.hitTimed (I)J\n"
                + "DUP2\nLSTORE 8\nL2I\nISTORE 10\nILOAD 10\nIFEQ L1\n"
                + "LDC 5\nINVOKESTATIC com/sun/btrace/instr/MethodTracker.getEndTs (I)J\n"
                + "LLOAD 8\nLSUB\nLSTORE 6\nDUP2\nLSTORE 11\n"
                + "ALOAD 0\nLLOAD 11\nLLOAD 6\nALOAD 1\nLLOAD 2\nALOAD 4\nALOAD 5\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$ArgsDuration2Sampled$args2 (Ljava/lang/Object;JJLjava/lang/String;J[Ljava/lang/String;[I)V\n"
                + "L1\nILOAD 10\nIFEQ L2\n"
                + "DUP2\nLSTORE 13\n"
                + "ALOAD 0\nLLOAD 13\nLLOAD 6\nALOAD 1\nLLOAD 2\nALOAD 4\nALOAD 5\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$ArgsDuration2Sampled$args (Ljava/lang/Object;JJLjava/lang/String;J[Ljava/lang/String;[I)V\n"
                + "L2\nL3\n"
        );
    }

    @Test
    public void methodEntryArgsDurationErr() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/ArgsDurationErr");

        checkTransformation("TRYCATCHBLOCK L0 L1 L1 java/lang/Throwable\n" +
            "LDC 0\n" +
            "LSTORE 6\n" +
            "INVOKESTATIC java/lang/System.nanoTime ()J\n" +
            "LSTORE 8\n" +
            "INVOKESTATIC java/lang/System.nanoTime ()J\n" +
            "LLOAD 8\n" +
            "LSUB\n" +
            "LSTORE 6\n" +
            "DUP\n" +
            "ASTORE 10\n" +
            "ALOAD 0\n" +
            "LLOAD 6\n" +
            "ALOAD 10\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$ArgsDurationErr$args (Ljava/lang/Object;JLjava/lang/Throwable;)V\n" +
            "ATHROW\n" +
            "MAXSTACK = 5\n" +
            "MAXLOCALS = 11"
        );
    }

    @Test
    public void methodEntryArgsDurationBoxedErr() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/ArgsDurationBoxedErr");
        checkTransformation("");
    }

    @Test
    public void methodEntryArgsDurationConstructorErr() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/ArgsDurationConstructorErr");
        checkTransformation("TRYCATCHBLOCK L0 L1 L1 java/lang/Throwable\n" +
            "L2\n" +
            "LINENUMBER 39 L2\n" +
            "LDC 0\n" +
            "LSTORE 1\n" +
            "INVOKESTATIC java/lang/System.nanoTime ()J\n" +
            "LSTORE 3\n" +
            "L0\n" +
            "INVOKESTATIC java/lang/System.nanoTime ()J\n" +
            "LLOAD 3\n" +
            "LSUB\n" +
            "LSTORE 1\n" +
            "DUP\n" +
            "ASTORE 5\n" +
            "ALOAD 0\n" +
            "LLOAD 1\n" +
            "ALOAD 5\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$ArgsDurationConstructorErr$args (Ljava/lang/Object;JLjava/lang/Throwable;)V\n" +
            "ATHROW\n" +
            "LOCALVARIABLE this Lresources/OnMethodTest; L2 L1 0\n" +
            "MAXSTACK = 5\n" +
            "MAXLOCALS = 6\n" +
            "TRYCATCHBLOCK L0 L1 L1 java/lang/Throwable\n" +
            "L2\n" +
            "LINENUMBER 40 L2\n" +
            "LDC 0\n" +
            "LSTORE 2\n" +
            "INVOKESTATIC java/lang/System.nanoTime ()J\n" +
            "LSTORE 4\n" +
            "L0\n" +
            "INVOKESTATIC java/lang/System.nanoTime ()J\n" +
            "LLOAD 4\n" +
            "LSUB\n" +
            "LSTORE 2\n" +
            "DUP\n" +
            "ASTORE 6\n" +
            "ALOAD 0\n" +
            "LLOAD 2\n" +
            "ALOAD 6\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$ArgsDurationConstructorErr$args (Ljava/lang/Object;JLjava/lang/Throwable;)V\n" +
            "ATHROW\n" +
            "LOCALVARIABLE this Lresources/OnMethodTest; L2 L1 0\n" +
            "LOCALVARIABLE a Ljava/lang/String; L2 L1 1\n" +
            "MAXSTACK = 5\n" +
            "MAXLOCALS = 7");
    }

    @Test
    // check for multiple timestamps
    public void methodEntryArgsDuration2Err() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/ArgsDuration2Err");
        checkTransformation("TRYCATCHBLOCK L0 L1 L1 java/lang/Throwable\n" +
            "TRYCATCHBLOCK L0 L2 L2 java/lang/Throwable\n" +
            "LDC 0\n" +
            "LSTORE 6\n" +
            "INVOKESTATIC java/lang/System.nanoTime ()J\n" +
            "LSTORE 8\n" +
            "INVOKESTATIC java/lang/System.nanoTime ()J\n" +
            "LLOAD 8\n" +
            "LSUB\n" +
            "LSTORE 6\n" +
            "DUP\n" +
            "ASTORE 10\n" +
            "ALOAD 0\n" +
            "LLOAD 6\n" +
            "ALOAD 10\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$ArgsDuration2Err$args2 (Ljava/lang/Object;JLjava/lang/Throwable;)V\n" +
            "ATHROW\n" +
            "L2\n" +
            "INVOKESTATIC java/lang/System.nanoTime ()J\n" +
            "LLOAD 8\n" +
            "LSUB\n" +
            "LSTORE 6\n" +
            "DUP\n" +
            "ASTORE 11\n" +
            "ALOAD 0\n" +
            "LLOAD 6\n" +
            "ALOAD 11\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$ArgsDuration2Err$args (Ljava/lang/Object;JLjava/lang/Throwable;)V\n" +
            "ATHROW\n" +
            "MAXSTACK = 5\n" +
            "MAXLOCALS = 12\n" +
            "TRYCATCHBLOCK L0 L1 L1 java/lang/Throwable\n" +
            "ATHROW\n" +
            "TRYCATCHBLOCK L0 L1 L1 java/lang/Throwable\n" +
            "LDC 0\n" +
            "LSTORE 6\n" +
            "INVOKESTATIC java/lang/System.nanoTime ()J\n" +
            "LSTORE 8\n" +
            "IFLE L2\n" +
            "L3\n" +
            "LINENUMBER 124 L3\n" +
            "L2\n" +
            "LINENUMBER 127 L2\n" +
            "IFLE L4\n" +
            "L5\n" +
            "LINENUMBER 128 L5\n" +
            "L4\n" +
            "LINENUMBER 132 L4\n" +
            "L6\n" +
            "LINENUMBER 133 L6\n" +
            "L1\n" +
            "INVOKESTATIC java/lang/System.nanoTime ()J\n" +
            "LLOAD 8\n" +
            "LSUB\n" +
            "LSTORE 6\n" +
            "DUP\n" +
            "ASTORE 10\n" +
            "ALOAD 0\n" +
            "LLOAD 6\n" +
            "ALOAD 10\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$ArgsDuration2Err$args2 (Ljava/lang/Object;JLjava/lang/Throwable;)V\n" +
            "ATHROW\n" +
            "LOCALVARIABLE this Lresources/OnMethodTest; L0 L1 0\n" +
            "LOCALVARIABLE a Ljava/lang/String; L0 L1 1\n" +
            "LOCALVARIABLE b J L0 L1 2\n" +
            "LOCALVARIABLE c [Ljava/lang/String; L0 L1 4\n" +
            "LOCALVARIABLE d [I L0 L1 5\n" +
            "MAXSTACK = 5\n" +
            "MAXLOCALS = 11");
    }

    @Test
    public void methodEntryAnytypeArgs() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/AnytypeArgs");
        checkTransformation("ALOAD 0\nICONST_4\nANEWARRAY java/lang/Object\nDUP\n"
                + "ICONST_0\nALOAD 1\nAASTORE\nDUP\n"
                + "ICONST_1\nLLOAD 2\nINVOKESTATIC java/lang/Long.valueOf (J)Ljava/lang/Long;\nAASTORE\nDUP\n"
                + "ICONST_2\nALOAD 4\nAASTORE\nDUP\n"
                + "ICONST_3\nALOAD 5\nAASTORE\nINVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$AnytypeArgs$args (Ljava/lang/Object;[Ljava/lang/Object;)V");
    }

    @Test
    public void methodEntryAnytypeArgsNoSelf() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/AnytypeArgsNoSelf");
        checkTransformation("ICONST_4\nANEWARRAY java/lang/Object\nDUP\nICONST_0\nALOAD 1\nAASTORE\n"
                + "DUP\nICONST_1\nLLOAD 2\nINVOKESTATIC java/lang/Long.valueOf (J)Ljava/lang/Long;\nAASTORE\n"
                + "DUP\nICONST_2\nALOAD 4\nAASTORE\nDUP\nICONST_3\nALOAD 5\nAASTORE\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$AnytypeArgsNoSelf$argsNoSelf ([Ljava/lang/Object;)V");
    }

    @Test
    public void methodEntryStaticArgs() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/StaticArgs");

        checkTransformation("ALOAD 0\nLLOAD 1\nALOAD 3\nALOAD 4\nINVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$StaticArgs$args (Ljava/lang/String;J[Ljava/lang/String;[I)V");
    }

    @Test
    public void methodEntryStaticArgsReturn() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/StaticArgsReturn");

        checkTransformation("DUP2\nLSTORE 5\nALOAD 0\nLLOAD 5\nLLOAD 1\nALOAD 3\nALOAD 4\nINVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$StaticArgsReturn$args (Ljava/lang/String;JJ[Ljava/lang/String;[I)V");
    }

    @Test
    public void methodEntryStaticArgsSelf() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/StaticArgsSelf");

        checkTransformation("");
    }

    @Test
    public void methodEntryStaticNoArgs() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/StaticNoArgs");

        checkTransformation("INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$StaticNoArgs$argsEmpty ()V");
    }

    @Test
    public void methodEntryStaticNoArgsSelf() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/StaticNoArgsSelf");

        checkTransformation("");
    }

    @Test
    public void methodCall() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/MethodCall");

        checkTransformation("LSTORE 4\nASTORE 6\nASTORE 7\nALOAD 0\nALOAD 6\nLLOAD 4\nALOAD 7\n"
                + "LDC \"resources/OnMethodTest.callTarget(Ljava/lang/String;J)J\"\nLDC \"resources/OnMethodTest\"\n"
                + "LDC \"callTopLevel\"\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$MethodCall$args (Ljava/lang/Object;Ljava/lang/String;JLjava/lang/Object;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V\n"
                + "ALOAD 7\nALOAD 6\nLLOAD 4");
    }

    @Test
    public void methodCallSampled() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/MethodCallSampled");

        checkTransformation("LDC 10\n" +
            "INVOKESTATIC com/sun/btrace/instr/MethodTracker.hit (I)Z\n" +
            "ISTORE 4\n" +
            "LSTORE 5\n" +
            "ASTORE 7\n" +
            "ASTORE 8\n" +
            "ILOAD 4\n" +
            "IFEQ L1\n" +
            "ALOAD 0\n" +
            "ALOAD 7\n" +
            "LLOAD 5\n" +
            "ALOAD 8\n" +
            "LDC \"resources/OnMethodTest.callTarget(Ljava/lang/String;J)J\"\n" +
            "LDC \"resources/OnMethodTest\"\n" +
            "LDC \"callTopLevel\"\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$MethodCallSampled$args (Ljava/lang/Object;Ljava/lang/String;JLjava/lang/Object;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V\n" +
            "L1\n" +
            "ALOAD 8\n" +
            "ALOAD 7\n" +
            "LLOAD 5\n" +
            "L2");
    }

    @Test
    public void methodCallSampledAdaptive() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/MethodCallSampledAdaptive");

        checkTransformation("LDC 10\n" +
            "INVOKESTATIC com/sun/btrace/instr/MethodTracker.hitAdaptive (I)Z\n" +
            "ISTORE 4\n" +
            "LSTORE 5\n" +
            "ASTORE 7\n" +
            "ASTORE 8\n" +
            "ILOAD 4\n" +
            "IFEQ L1\n" +
            "ALOAD 0\n" +
            "ALOAD 7\n" +
            "LLOAD 5\n" +
            "ALOAD 8\n" +
            "LDC \"resources/OnMethodTest.callTarget(Ljava/lang/String;J)J\"\n" +
            "LDC \"resources/OnMethodTest\"\n" +
            "LDC \"callTopLevel\"\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$MethodCallSampledAdaptive$args (Ljava/lang/Object;Ljava/lang/String;JLjava/lang/Object;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V\n" +
            "L1\n" +
            "ALOAD 8\n" +
            "ALOAD 7\n" +
            "LLOAD 5\n" +
            "ILOAD 4\n" +
            "IFEQ L2\n" +
            "LDC 9\n" +
            "INVOKESTATIC com/sun/btrace/instr/MethodTracker.updateEndTs (I)V\n" +
            "L2\n" +
            "L3");
    }

    @Test
    public void methodCallNoArgs() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/MethodCallNoArgs");

        checkTransformation("INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$MethodCallNoArgs$args ()V");
    }

    @Test
    public void methodCallReturn() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/MethodCallReturn");

        checkTransformation("LSTORE 4\nASTORE 6\nASTORE 7\nALOAD 7\nALOAD 6\nLLOAD 4\n"
                + "LSTORE 8\nLLOAD 8\nALOAD 6\nLLOAD 4\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$MethodCallReturn$args (JLjava/lang/String;J)V\n"
                + "LLOAD 8");
    }

    @Test
    public void methodCallReturnAugmented() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/MethodCallReturnAugmented", true);

        checkTransformation("LSTORE 4\n" +
            "ASTORE 6\n" +
            "ASTORE 7\n" +
            "ALOAD 7\n" +
            "ALOAD 6\n" +
            "LLOAD 4\n" +
            "LSTORE 8\n" +
            "ALOAD 6\n" +
            "LLOAD 4\n" +
            "LLOAD 8\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$MethodCallReturnAugmented$args (Ljava/lang/String;JJ)J\n" +
            "MAXLOCALS = 10\n" +
            "\n" +
            "// access flags 0xA\n" +
            "private static $btrace$traces$onmethod$MethodCallReturnAugmented$args(Ljava/lang/String;JJ)J\n" +
            "@Lcom/sun/btrace/annotations/OnMethod;(clazz=\"/.*\\\\.OnMethodTest/\", method=\"callTopLevel\", location=@Lcom/sun/btrace/annotations/Location;(value=Lcom/sun/btrace/annotations/Kind;.CALL, clazz=\"/.*\\\\.OnMethodTest/\", method=\"callTarget\", where=Lcom/sun/btrace/annotations/Where;.AFTER))\n" +
            "@Lcom/sun/btrace/annotations/Return;() // parameter 2\n" +
            "TRYCATCHBLOCK L0 L1 L1 java/lang/Throwable\n" +
            "GETSTATIC traces/onmethod/MethodCallReturnAugmented.runtime : Lcom/sun/btrace/BTraceRuntime;\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.enter (Lcom/sun/btrace/BTraceRuntime;)Z\n" +
            "IFNE L0\n" +
            "LLOAD 3\n" +
            "LRETURN\n" +
            "L0\n" +
            "LDC \"args\"\n" +
            "INVOKESTATIC com/sun/btrace/BTraceUtils.println (Ljava/lang/Object;)V\n" +
            "LCONST_1\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "LRETURN\n" +
            "L1\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.handleException (Ljava/lang/Throwable;)V\n" +
            "LLOAD 3\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "LRETURN\n" +
            "MAXSTACK = 2\n" +
            "MAXLOCALS = 5"
        );
    }

    @Test
    public void methodCallReturnAugmented1() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/MethodCallReturnAugmented1", true);

        checkTransformation("LSTORE 4\n" +
            "ASTORE 6\n" +
            "ASTORE 7\n" +
            "ALOAD 7\n" +
            "ALOAD 6\n" +
            "LLOAD 4\n" +
            "LSTORE 8\n" +
            "LLOAD 8\n" +
            "ALOAD 6\n" +
            "LLOAD 4\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$MethodCallReturnAugmented1$args (JLjava/lang/String;J)J\n" +
            "MAXLOCALS = 10\n" +
            "\n" +
            "// access flags 0xA\n" +
            "private static $btrace$traces$onmethod$MethodCallReturnAugmented1$args(JLjava/lang/String;J)J\n" +
            "@Lcom/sun/btrace/annotations/OnMethod;(clazz=\"/.*\\\\.OnMethodTest/\", method=\"callTopLevel\", location=@Lcom/sun/btrace/annotations/Location;(value=Lcom/sun/btrace/annotations/Kind;.CALL, clazz=\"/.*\\\\.OnMethodTest/\", method=\"callTarget\", where=Lcom/sun/btrace/annotations/Where;.AFTER))\n" +
            "@Lcom/sun/btrace/annotations/Return;() // parameter 0\n" +
            "TRYCATCHBLOCK L0 L1 L1 java/lang/Throwable\n" +
            "GETSTATIC traces/onmethod/MethodCallReturnAugmented1.runtime : Lcom/sun/btrace/BTraceRuntime;\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.enter (Lcom/sun/btrace/BTraceRuntime;)Z\n" +
            "IFNE L0\n" +
            "LLOAD 0\n" +
            "LRETURN\n" +
            "L0\n" +
            "LDC \"args\"\n" +
            "INVOKESTATIC com/sun/btrace/BTraceUtils.println (Ljava/lang/Object;)V\n" +
            "LLOAD 0\n" +
            "LCONST_1\n" +
            "LADD\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "LRETURN\n" +
            "L1\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.handleException (Ljava/lang/Throwable;)V\n" +
            "LLOAD 0\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "LRETURN\n" +
            "MAXSTACK = 4\n" +
            "MAXLOCALS = 5"
        );
    }

    @Test
    public void methodCallDuration() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/MethodCallDuration");

        checkTransformation("LDC 0\n" +
            "LSTORE 4\n" +
            "INVOKESTATIC java/lang/System.nanoTime ()J\n" +
            "LSTORE 6\n" +
            "LSTORE 8\n" +
            "ASTORE 10\n" +
            "ASTORE 11\n" +
            "ALOAD 11\n" +
            "ALOAD 10\n" +
            "LLOAD 8\n" +
            "INVOKESTATIC java/lang/System.nanoTime ()J\n" +
            "LLOAD 6\n" +
            "LSUB\n" +
            "LSTORE 4\n" +
            "LSTORE 12\n" +
            "LLOAD 12\n" +
            "LLOAD 4\n" +
            "ALOAD 10\n" +
            "LLOAD 8\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$MethodCallDuration$args (JJLjava/lang/String;J)V\n" +
            "LLOAD 12\n" +
            "MAXSTACK = 7\n" +
            "MAXLOCALS = 14");
    }

    @Test
    public void methodCallDurationSampled() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/MethodCallDurationSampled");

        checkTransformation("LDC 0\n" +
            "LSTORE 4\n" +
            "LDC 10\n" +
            "INVOKESTATIC com/sun/btrace/instr/MethodTracker.hitTimed (I)J\n" +
            "DUP2\n" +
            "LSTORE 6\n" +
            "L2I\n" +
            "ISTORE 8\n" +
            "LSTORE 9\n" +
            "ASTORE 11\n" +
            "ASTORE 12\n" +
            "ALOAD 12\n" +
            "ALOAD 11\n" +
            "LLOAD 9\n" +
            "ILOAD 8\n" +
            "IFEQ L1\n" +
            "LDC 10\n" +
            "INVOKESTATIC com/sun/btrace/instr/MethodTracker.getEndTs (I)J\n" +
            "LLOAD 6\n" +
            "LSUB\n" +
            "LSTORE 4\n" +
            "LSTORE 13\n" +
            "LLOAD 13\n" +
            "LLOAD 4\n" +
            "ALOAD 11\n" +
            "LLOAD 9\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$MethodCallDurationSampled$args (JJLjava/lang/String;J)V\n" +
            "LLOAD 13\n" +
            "L1\n" +
            "L2");
    }

    @Test
    public void methodCallDurationSampledMulti() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/MethodCallDurationSampledMulti");

        checkTransformation("LDC 0\n" +
            "LSTORE 4\n" +
            "LDC 20\n" +
            "INVOKESTATIC com/sun/btrace/instr/MethodTracker.hitTimed (I)J\n" +
            "DUP2\n" +
            "LSTORE 6\n" +
            "L2I\n" +
            "ISTORE 8\n" +
            "LSTORE 9\n" +
            "ASTORE 11\n" +
            "ASTORE 12\n" +
            "ALOAD 12\n" +
            "ALOAD 11\n" +
            "LLOAD 9\n" +
            "ILOAD 8\n" +
            "IFEQ L1\n" +
            "LDC 20\n" +
            "INVOKESTATIC com/sun/btrace/instr/MethodTracker.getEndTs (I)J\n" +
            "LLOAD 6\n" +
            "LSUB\n" +
            "LSTORE 4\n" +
            "LSTORE 13\n" +
            "LLOAD 13\n" +
            "LLOAD 4\n" +
            "ALOAD 11\n" +
            "LLOAD 9\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$MethodCallDurationSampledMulti$args (JJLjava/lang/String;J)V\n" +
            "LLOAD 13\n" +
            "L1\n" +
            "LDC 21\n" +
            "INVOKESTATIC com/sun/btrace/instr/MethodTracker.hitTimed (I)J\n" +
            "DUP2\n" +
            "LSTORE 15\n" +
            "L2I\n" +
            "ISTORE 17\n" +
            "LSTORE 18\n" +
            "ASTORE 20\n" +
            "ALOAD 20\n" +
            "LLOAD 18\n" +
            "ILOAD 17\n" +
            "IFEQ L2\n" +
            "LDC 21\n" +
            "INVOKESTATIC com/sun/btrace/instr/MethodTracker.getEndTs (I)J\n" +
            "LLOAD 15\n" +
            "LSUB\n" +
            "LSTORE 21\n" +
            "LLOAD 21\n" +
            "ALOAD 20\n" +
            "LLOAD 18\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$MethodCallDurationSampledMulti$args (JJLjava/lang/String;J)V\n" +
            "LLOAD 21\n" +
            "L2\n" +
            "LADD\n" +
            "LSTORE 23\n" +
            "L3\n" +
            "LINENUMBER 115 L3\n" +
            "LLOAD 23\n" +
            "L4");
    }

    // multiple instrumentation of a call site is not handled well
//    @Test
//    public void methodCallDuration2() throws Exception {
//        loadTargetClass("OnMethodTest");
//        transform("onmethod/MethodCallDuration2");
//
//        checkTransformation("LSTORE 4\nASTORE 6\nASTORE 7\n"
//                + "INVOKESTATIC java/lang/System.nanoTime ()J\n"
//                + "LSTORE 8\nALOAD 7\nALOAD 6\nLLOAD 4\nLSTORE 10\n"
//                + "INVOKESTATIC java/lang/System.nanoTime ()J\n"
//                + "LSTORE 12\nLLOAD 10\nLLOAD 12\nLLOAD 8\nLSUB\n"
//                + "ALOAD 6\nLLOAD 4\n"
//                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$MethodCallDuration$args (JJLjava/lang/String;J)V\n"
//                + "LLOAD 10\nMAXSTACK = 7\nMAXLOCALS = 14\n");
//    }


    @Test
    public void methodCallStatic() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/MethodCallStatic");

        checkTransformation("LSTORE 4\nASTORE 6\nALOAD 0\nALOAD 6\nLLOAD 4\n"
                + "LDC \"resources/OnMethodTest.callTargetStatic(Ljava/lang/String;J)J\"\nLDC \"resources/OnMethodTest\"\n"
                + "LDC \"callTopLevel\"\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$MethodCallStatic$args (Ljava/lang/Object;Ljava/lang/String;JLjava/lang/String;Ljava/lang/String;Ljava/lang/String;)V\n"
                + "ALOAD 6\nLLOAD 4");
    }

    @Test
    public void staticMethodCall() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/StaticMethodCall");

        checkTransformation("LSTORE 4\nASTORE 6\nASTORE 7\nALOAD 6\nLLOAD 4\nALOAD 7\n"
                + "LDC \"resources/OnMethodTest.callTarget(Ljava/lang/String;J)J\"\nLDC \"resources/OnMethodTest\"\n"
                + "LDC \"callTopLevelStatic\"\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$StaticMethodCall$args (Ljava/lang/String;JLjava/lang/Object;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V\n"
                + "ALOAD 7\nALOAD 6\nLLOAD 4");
    }

    @Test
    public void staticMethodCallStatic() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/StaticMethodCallStatic");

        checkTransformation("LSTORE 4\nASTORE 6\nALOAD 6\nLLOAD 4\n"
                + "LDC \"resources/OnMethodTest.callTargetStatic(Ljava/lang/String;J)J\"\nLDC \"resources/OnMethodTest\"\n"
                + "LDC \"callTopLevelStatic\"\n"
                + "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$StaticMethodCallStatic$args (Ljava/lang/String;JLjava/lang/String;Ljava/lang/String;Ljava/lang/String;)V\n"
                + "ALOAD 6\nLLOAD 4");
    }

    @Test
    public void methodEntryNoArgsEntryReturn() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/NoArgsEntryReturn");

        checkTransformation("ALOAD 0\nINVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$NoArgsEntryReturn$argsEmptyEntry (Ljava/lang/Object;)V\n"
                + "ALOAD 0\nINVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$NoArgsEntryReturn$argsEmptyReturn (Ljava/lang/Object;)V");
    }

    @Test
    public void servicesTest() throws Exception {
        // a sanity test for the runtime verifier accepting the services method calls
        loadTargetClass("OnMethodTest");
        transform("ServicesTest");

        checkTransformation(
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$ServicesTest$testSimpleService ()V\n" +
            "LDC \"resources.OnMethodTest\"\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$ServicesTest$testFieldInjection (Ljava/lang/String;)V\n" +
            "MAXSTACK = 1\n" +
            "ALOAD 1\n" +
            "LLOAD 2\n" +
            "ALOAD 4\n" +
            "ALOAD 5\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$ServicesTest$testRuntimeService (Ljava/lang/String;J[Ljava/lang/String;[I)V\n" +
            "MAXSTACK = 5\n" +
            "ALOAD 0\n" +
            "LLOAD 1\n" +
            "ALOAD 3\n" +
            "ALOAD 4\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$ServicesTest$testSingletonService (Ljava/lang/String;J[Ljava/lang/String;[I)V\n" +
            "MAXSTACK = 5\n" +
            "\n" +
            "// access flags 0xA\n" +
            "private static $btrace$traces$ServicesTest$testRuntimeService(Ljava/lang/String;J[Ljava/lang/String;[I)V\n" +
            "@Lcom/sun/btrace/annotations/OnMethod;(clazz=\"resources.OnMethodTest\", method=\"args\")\n" +
            "TRYCATCHBLOCK L0 L1 L1 java/lang/Throwable\n" +
            "GETSTATIC traces/ServicesTest.runtime : Lcom/sun/btrace/BTraceRuntime;\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.enter (Lcom/sun/btrace/BTraceRuntime;)Z\n" +
            "IFNE L0\n" +
            "RETURN\n" +
            "L0\n" +
            "NEW resources/services/DummyRuntimService\n" +
            "DUP\n" +
            "GETSTATIC traces/ServicesTest.runtime : Lcom/sun/btrace/BTraceRuntime;\n" +
            "INVOKESPECIAL resources/services/DummyRuntimService.<init> (Lcom/sun/btrace/BTraceRuntime;)V\n" +
            "ASTORE 5\n" +
            "ALOAD 5\n" +
            "BIPUSH 10\n" +
            "LDC \"hello\"\n" +
            "INVOKEVIRTUAL resources/services/DummyRuntimService.doit (ILjava/lang/String;)V\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "RETURN\n" +
            "L1\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.handleException (Ljava/lang/Throwable;)V\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "RETURN\n" +
            "MAXSTACK = 3\n" +
            "MAXLOCALS = 6\n" +
            "\n" +
            "// access flags 0xA\n" +
            "private static $btrace$traces$ServicesTest$testSimpleService()V\n" +
            "@Lcom/sun/btrace/annotations/OnMethod;(clazz=\"resources.OnMethodTest\", method=\"noargs\")\n" +
            "TRYCATCHBLOCK L0 L1 L1 java/lang/Throwable\n" +
            "GETSTATIC traces/ServicesTest.runtime : Lcom/sun/btrace/BTraceRuntime;\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.enter (Lcom/sun/btrace/BTraceRuntime;)Z\n" +
            "IFNE L0\n" +
            "RETURN\n" +
            "L0\n" +
            "NEW resources/services/DummySimpleService\n" +
            "DUP\n" +
            "INVOKESPECIAL resources/services/DummySimpleService.<init> ()V\n" +
            "ASTORE 0\n" +
            "ALOAD 0\n" +
            "LDC \"hello\"\n" +
            "BIPUSH 10\n" +
            "INVOKEVIRTUAL resources/services/DummySimpleService.doit (Ljava/lang/String;I)V\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "RETURN\n" +
            "L1\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.handleException (Ljava/lang/Throwable;)V\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "RETURN\n" +
            "MAXSTACK = 3\n" +
            "MAXLOCALS = 1\n" +
            "\n" +
            "// access flags 0xA\n" +
            "private static $btrace$traces$ServicesTest$testSingletonService(Ljava/lang/String;J[Ljava/lang/String;[I)V\n" +
            "@Lcom/sun/btrace/annotations/OnMethod;(clazz=\"resources.OnMethodTest\", method=\"args$static\")\n" +
            "TRYCATCHBLOCK L0 L1 L1 java/lang/Throwable\n" +
            "GETSTATIC traces/ServicesTest.runtime : Lcom/sun/btrace/BTraceRuntime;\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.enter (Lcom/sun/btrace/BTraceRuntime;)Z\n" +
            "IFNE L0\n" +
            "RETURN\n" +
            "L0\n" +
            "NEW resources/services/DummySimpleService\n" +
            "DUP\n" +
            "LDC \"getInstance\"\n" +
            "INVOKESPECIAL resources/services/DummySimpleService.<init> (Ljava/lang/String;)V\n" +
            "ASTORE 5\n" +
            "ALOAD 5\n" +
            "LDC \"hello\"\n" +
            "BIPUSH 10\n" +
            "INVOKEVIRTUAL resources/services/DummySimpleService.doit (Ljava/lang/String;I)V\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "RETURN\n" +
            "L1\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.handleException (Ljava/lang/Throwable;)V\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "RETURN\n" +
            "MAXSTACK = 3\n" +
            "MAXLOCALS = 6\n" +
            "\n" +
            "// access flags 0xA\n" +
            "private static $btrace$traces$ServicesTest$testFieldInjection(Ljava/lang/String;)V\n" +
            "@Lcom/sun/btrace/annotations/OnMethod;(clazz=\"resources.OnMethodTest\", method=\"noargs$static\")\n" +
            "@Lcom/sun/btrace/annotations/ProbeClassName;() // parameter 0\n" +
            "TRYCATCHBLOCK L0 L1 L1 java/lang/Throwable\n" +
            "GETSTATIC traces/ServicesTest.runtime : Lcom/sun/btrace/BTraceRuntime;\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.enter (Lcom/sun/btrace/BTraceRuntime;)Z\n" +
            "IFNE L0\n" +
            "RETURN\n" +
            "L0\n" +
            "NEW resources/services/DummyRuntimService\n" +
            "DUP\n" +
            "DUP\n" +
            "GETSTATIC traces/ServicesTest.runtime : Lcom/sun/btrace/BTraceRuntime;\n" +
            "INVOKESPECIAL resources/services/DummyRuntimService.<init> (Lcom/sun/btrace/BTraceRuntime;)V\n" +
            "ASTORE 1\n" +
            "BIPUSH 10\n" +
            "LDC \"hey\"\n" +
            "INVOKEVIRTUAL resources/services/DummyRuntimService.doit (ILjava/lang/String;)V\n" +
            "ALOAD 1\n" +
            "BIPUSH 20\n" +
            "LDC \"ho\"\n" +
            "INVOKEVIRTUAL resources/services/DummyRuntimService.doit (ILjava/lang/String;)V\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "RETURN\n" +
            "L1\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.handleException (Ljava/lang/Throwable;)V\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "RETURN\n" +
            "MAXSTACK = 4\n" +
            "MAXLOCALS = 2"
        );
    }

    @Test
    public void unsafeTest() throws Exception {
        loadTargetClass("OnMethodTest");
        System.err.println(asmify(originalBC));
        transform("onmethod/ArgsUnsafe", true);
        System.err.println("\n" + asmify(transformedBC));
    }

    @Test
    public void tlsTest() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("TLSTest");

        checkTrace(
            "public static Lcom/sun/btrace/BTraceRuntime; runtime\n" +
            "// access flags 0x19\n" +
            "// signature Ljava/lang/ThreadLocal<Ljava/util/Deque;>;\n" +
            "// declaration: java.lang.ThreadLocal<java.util.Deque>\n" +
            "public final static Ljava/lang/ThreadLocal; entryTimes\n" +
            "// access flags 0x19\n" +
            "// signature Ljava/lang/ThreadLocal<Ljava/lang/String;>;\n" +
            "// declaration: java.lang.ThreadLocal<java.lang.String>\n" +
            "public final static Ljava/lang/ThreadLocal; name\n" +
            "// access flags 0x19\n" +
            "// signature Ljava/lang/ThreadLocal<Ljava/lang/Integer;>;\n" +
            "// declaration: java.lang.ThreadLocal<java.lang.Integer>\n" +
            "public final static Ljava/lang/ThreadLocal; x\n" +
            "// access flags 0x19\n" +
            "// signature Ljava/lang/ThreadLocal<Ljava/lang/Double;>;\n" +
            "// declaration: java.lang.ThreadLocal<java.lang.Double>\n" +
            "public final static Ljava/lang/ThreadLocal; y\n" +
            "// signature Ljava/lang/ThreadLocal<Ljava/lang/Long;>;\n" +
            "// declaration: java.lang.ThreadLocal<java.lang.Long>\n" +
            "public final static Ljava/lang/ThreadLocal; z\n" +
            "TRYCATCHBLOCK L0 L1 L1 java/lang/Throwable\n" +
            "GETSTATIC traces/TLSTest.runtime : Lcom/sun/btrace/BTraceRuntime;\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.enter (Lcom/sun/btrace/BTraceRuntime;)Z\n" +
            "IFNE L0\n" +
            "RETURN\n" +
            "L0\n" +
            "GETSTATIC traces/TLSTest.entryTimes : Ljava/lang/ThreadLocal;\n" +
            "INVOKEVIRTUAL java/lang/ThreadLocal.get ()Ljava/lang/Object;\n" +
            "CHECKCAST java/util/Deque\n" +
            "GETSTATIC traces/TLSTest.name : Ljava/lang/ThreadLocal;\n" +
            "SWAP\n" +
            "INVOKEVIRTUAL java/lang/ThreadLocal.set (Ljava/lang/Object;)V\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "RETURN\n" +
            "L1\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.handleException (Ljava/lang/Throwable;)V\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "TRYCATCHBLOCK L0 L1 L1 java/lang/Throwable\n" +
            "L0\n" +
            "LDC Ltraces/TLSTest;.class\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.forClass (Ljava/lang/Class;)Lcom/sun/btrace/BTraceRuntime;\n" +
            "PUTSTATIC traces/TLSTest.runtime : Lcom/sun/btrace/BTraceRuntime;\n" +
            "GETSTATIC traces/TLSTest.runtime : Lcom/sun/btrace/BTraceRuntime;\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.enter (Lcom/sun/btrace/BTraceRuntime;)Z\n" +
            "IFNE L2\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "RETURN\n" +
            "L2\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.newThreadLocal (Ljava/lang/Object;)Ljava/lang/ThreadLocal;\n" +
            "PUTSTATIC traces/TLSTest.entryTimes : Ljava/lang/ThreadLocal;\n" +
            "INVOKESTATIC java/lang/Integer.valueOf (I)Ljava/lang/Integer;\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.newThreadLocal (Ljava/lang/Object;)Ljava/lang/ThreadLocal;\n" +
            "PUTSTATIC traces/TLSTest.x : Ljava/lang/ThreadLocal;\n" +
            "ACONST_NULL\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.newThreadLocal (Ljava/lang/Object;)Ljava/lang/ThreadLocal;\n" +
            "PUTSTATIC traces/TLSTest.name : Ljava/lang/ThreadLocal;\n" +
            "DCONST_0\n" +
            "INVOKESTATIC java/lang/Double.valueOf (D)Ljava/lang/Double;\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.newThreadLocal (Ljava/lang/Object;)Ljava/lang/ThreadLocal;\n" +
            "PUTSTATIC traces/TLSTest.y : Ljava/lang/ThreadLocal;\n" +
            "LDC 10\n" +
            "INVOKESTATIC java/lang/Long.valueOf (J)Ljava/lang/Long;\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.newThreadLocal (Ljava/lang/Object;)Ljava/lang/ThreadLocal;\n" +
            "PUTSTATIC traces/TLSTest.z : Ljava/lang/ThreadLocal;\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.start ()V\n" +
            "RETURN\n" +
            "L1\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.handleException (Ljava/lang/Throwable;)V\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "MAXSTACK = 2"
        );

        checkTransformation(
            "ALOAD 1\n" +
            "LLOAD 2\n" +
            "ALOAD 4\n" +
            "ALOAD 5\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$TLSTest$testArgs (Ljava/lang/String;J[Ljava/lang/String;[I)V\n" +
            "MAXSTACK = 5\n" +
            "\n" +
            "// access flags 0xA\n" +
            "private static $btrace$traces$TLSTest$testArgs(Ljava/lang/String;J[Ljava/lang/String;[I)V\n" +
            "@Lcom/sun/btrace/annotations/OnMethod;(clazz=\"resources.OnMethodTest\", method=\"args\")\n" +
            "TRYCATCHBLOCK L0 L1 L1 java/lang/Throwable\n" +
            "GETSTATIC traces/TLSTest.runtime : Lcom/sun/btrace/BTraceRuntime;\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.enter (Lcom/sun/btrace/BTraceRuntime;)Z\n" +
            "IFNE L0\n" +
            "RETURN\n" +
            "L0\n" +
            "GETSTATIC traces/TLSTest.entryTimes : Ljava/lang/ThreadLocal;\n" +
            "INVOKEVIRTUAL java/lang/ThreadLocal.get ()Ljava/lang/Object;\n" +
            "CHECKCAST java/util/Deque\n" +
            "LLOAD 1\n" +
            "INVOKESTATIC java/lang/Long.valueOf (J)Ljava/lang/Long;\n" +
            "INVOKESTATIC com/sun/btrace/BTraceUtils.push (Ljava/util/Deque;Ljava/lang/Object;)V\n" +
            "ALOAD 0\n" +
            "GETSTATIC traces/TLSTest.name : Ljava/lang/ThreadLocal;\n" +
            "SWAP\n" +
            "INVOKEVIRTUAL java/lang/ThreadLocal.set (Ljava/lang/Object;)V\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "RETURN\n" +
            "L1\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.handleException (Ljava/lang/Throwable;)V\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "RETURN\n" +
            "MAXSTACK = 3\n" +
            "MAXLOCALS = 5"
        );
    }

    @Test
    public void exportTest() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("ExportTest");

        checkTrace(
            "public static Lcom/sun/btrace/BTraceRuntime; runtime\n" +
            "// access flags 0x19\n" +
            "public final static Ljava/util/Deque; entryTimes\n" +
            "// access flags 0x19\n" +
            "public final static Ljava/lang/String; name\n" +
            "// access flags 0x19\n" +
            "public final static I x\n" +
            "// access flags 0x19\n" +
            "public final static D y\n" +
            "public final static J z\n" +
            "TRYCATCHBLOCK L0 L1 L1 java/lang/Throwable\n" +
            "GETSTATIC traces/ExportTest.runtime : Lcom/sun/btrace/BTraceRuntime;\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.enter (Lcom/sun/btrace/BTraceRuntime;)Z\n" +
            "IFNE L0\n" +
            "RETURN\n" +
            "L0\n" +
            "ACONST_NULL\n" +
            "LDC \"btrace.traces/ExportTest.name\"\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.putPerfString (Ljava/lang/String;Ljava/lang/String;)V\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "RETURN\n" +
            "L1\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.handleException (Ljava/lang/Throwable;)V\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "TRYCATCHBLOCK L0 L1 L1 java/lang/Throwable\n" +
            "L0\n" +
            "LDC Ltraces/ExportTest;.class\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.forClass (Ljava/lang/Class;)Lcom/sun/btrace/BTraceRuntime;\n" +
            "PUTSTATIC traces/ExportTest.runtime : Lcom/sun/btrace/BTraceRuntime;\n" +
            "GETSTATIC traces/ExportTest.runtime : Lcom/sun/btrace/BTraceRuntime;\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.enter (Lcom/sun/btrace/BTraceRuntime;)Z\n" +
            "IFNE L2\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "RETURN\n" +
            "L2\n" +
            "LDC \"btrace.traces/ExportTest.entryTimes\"\n" +
            "LDC \"Ljava/util/Deque;\"\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.newPerfCounter (Ljava/lang/Object;Ljava/lang/String;Ljava/lang/String;)V\n" +
            "INVOKESTATIC java/lang/Integer.valueOf (I)Ljava/lang/Integer;\n" +
            "LDC \"btrace.traces/ExportTest.x\"\n" +
            "LDC \"I\"\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.newPerfCounter (Ljava/lang/Object;Ljava/lang/String;Ljava/lang/String;)V\n" +
            "ACONST_NULL\n" +
            "LDC \"btrace.traces/ExportTest.name\"\n" +
            "LDC \"Ljava/lang/String;\"\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.newPerfCounter (Ljava/lang/Object;Ljava/lang/String;Ljava/lang/String;)V\n" +
            "DCONST_0\n" +
            "INVOKESTATIC java/lang/Double.valueOf (D)Ljava/lang/Double;\n" +
            "LDC \"btrace.traces/ExportTest.y\"\n" +
            "LDC \"D\"\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.newPerfCounter (Ljava/lang/Object;Ljava/lang/String;Ljava/lang/String;)V\n" +
            "LDC 10\n" +
            "INVOKESTATIC java/lang/Long.valueOf (J)Ljava/lang/Long;\n" +
            "LDC \"btrace.traces/ExportTest.z\"\n" +
            "LDC \"J\"\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.newPerfCounter (Ljava/lang/Object;Ljava/lang/String;Ljava/lang/String;)V\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.start ()V\n" +
            "RETURN\n" +
            "L1\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.handleException (Ljava/lang/Throwable;)V\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "MAXSTACK = 3"
        );

        checkTransformation(
            "ALOAD 1\n" +
            "LLOAD 2\n" +
            "ALOAD 4\n" +
            "ALOAD 5\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$ExportTest$testArgs (Ljava/lang/String;J[Ljava/lang/String;[I)V\n" +
            "MAXSTACK = 5\n" +
            "\n" +
            "// access flags 0xA\n" +
            "private static $btrace$traces$ExportTest$testArgs(Ljava/lang/String;J[Ljava/lang/String;[I)V\n" +
            "@Lcom/sun/btrace/annotations/OnMethod;(clazz=\"resources.OnMethodTest\", method=\"args\")\n" +
            "TRYCATCHBLOCK L0 L1 L1 java/lang/Throwable\n" +
            "GETSTATIC traces/ExportTest.runtime : Lcom/sun/btrace/BTraceRuntime;\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.enter (Lcom/sun/btrace/BTraceRuntime;)Z\n" +
            "IFNE L0\n" +
            "RETURN\n" +
            "L0\n" +
            "ACONST_NULL\n" +
            "LLOAD 1\n" +
            "INVOKESTATIC java/lang/Long.valueOf (J)Ljava/lang/Long;\n" +
            "INVOKESTATIC com/sun/btrace/BTraceUtils.push (Ljava/util/Deque;Ljava/lang/Object;)V\n" +
            "ALOAD 0\n" +
            "LDC \"btrace.traces/ExportTest.name\"\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.putPerfString (Ljava/lang/String;Ljava/lang/String;)V\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "RETURN\n" +
            "L1\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.handleException (Ljava/lang/Throwable;)V\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "RETURN\n" +
            "MAXSTACK = 3\n" +
            "MAXLOCALS = 5"
        );
    }

    @Test
    public void onprobeTest() throws Exception {
        loadTargetClass("OnMethodTest");
        System.err.println(asmify(originalBC));
        transform("OnProbeTest", false);
        System.err.println("\n" + asmify(transformedBC));
    }

    @Test
    public void onTimerTest() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("OnTimerTest", false);

        checkTrace(
            "public static Lcom/sun/btrace/BTraceRuntime; runtime\n" +
            "TRYCATCHBLOCK L0 L1 L1 java/lang/Throwable\n" +
            "GETSTATIC traces/OnTimerTest.runtime : Lcom/sun/btrace/BTraceRuntime;\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.enter (Lcom/sun/btrace/BTraceRuntime;)Z\n" +
            "IFNE L0\n" +
            "RETURN\n" +
            "L0\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "L1\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.handleException (Ljava/lang/Throwable;)V\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "RETURN\n" +
            "TRYCATCHBLOCK L0 L1 L1 java/lang/Throwable\n" +
            "L0\n" +
            "LDC Ltraces/OnTimerTest;.class\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.forClass (Ljava/lang/Class;)Lcom/sun/btrace/BTraceRuntime;\n" +
            "PUTSTATIC traces/OnTimerTest.runtime : Lcom/sun/btrace/BTraceRuntime;\n" +
            "GETSTATIC traces/OnTimerTest.runtime : Lcom/sun/btrace/BTraceRuntime;\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.enter (Lcom/sun/btrace/BTraceRuntime;)Z\n" +
            "IFNE L2\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "RETURN\n" +
            "L2\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.start ()V\n" +
            "L1\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.handleException (Ljava/lang/Throwable;)V\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "RETURN"
        );
    }

    @Test
    public void nativeWithReturnTracingTest() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/NativeWithReturn", false);

        checkTransformation(
            "// access flags 0x1\n" +
            "public nativeWithReturn(ILjava/lang/String;[J[Ljava/lang/Object;)J\n" +
            "ALOAD 0\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$NativeWithReturn$nMethod (Ljava/lang/Object;)V\n" +
            "ALOAD 0\n" +
            "ILOAD 1\n" +
            "ALOAD 2\n" +
            "ALOAD 3\n" +
            "ALOAD 4\n" +
            "INVOKESPECIAL resources/OnMethodTest.$btrace$native$nativeWithReturn (ILjava/lang/String;[J[Ljava/lang/Object;)J\n" +
            "LRETURN\n" +
            "MAXSTACK = 5\n" +
            "MAXLOCALS = 5\n" +
            "\n" +
            "public native $btrace$native$nativeWithReturn(ILjava/lang/String;[J[Ljava/lang/Object;)J\n" +
            "\n" +
            "// access flags 0xA\n" +
            "private static $btrace$traces$onmethod$NativeWithReturn$nMethod(Ljava/lang/Object;)V\n" +
            "@Lcom/sun/btrace/annotations/OnMethod;(clazz=\"/.*\\\\.OnMethodTest/\", method=\"nativeWithReturn\")\n" +
            "@Lcom/sun/btrace/annotations/Self;() // parameter 0\n" +
            "TRYCATCHBLOCK L0 L1 L1 java/lang/Throwable\n" +
            "GETSTATIC traces/onmethod/NativeWithReturn.runtime : Lcom/sun/btrace/BTraceRuntime;\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.enter (Lcom/sun/btrace/BTraceRuntime;)Z\n" +
            "IFNE L0\n" +
            "RETURN\n" +
            "L0\n" +
            "LDC \"args\"\n" +
            "INVOKESTATIC com/sun/btrace/BTraceUtils.println (Ljava/lang/Object;)V\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "RETURN\n" +
            "L1\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.handleException (Ljava/lang/Throwable;)V\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "RETURN\n" +
            "MAXSTACK = 1\n" +
            "MAXLOCALS = 1"
        );
    }

    @Test
    public void nativeWithoutReturnTracingTest() throws Exception {
        loadTargetClass("OnMethodTest");
        transform("onmethod/NativeWithoutReturn", false);

        checkTransformation(
            "// access flags 0x1\n" +
            "public nativeWithoutReturn(ILjava/lang/String;[J[Ljava/lang/Object;)V\n" +
            "ALOAD 0\n" +
            "INVOKESTATIC resources/OnMethodTest.$btrace$traces$onmethod$NativeWithoutReturn$nMethod (Ljava/lang/Object;)V\n" +
            "ALOAD 0\n" +
            "ILOAD 1\n" +
            "ALOAD 2\n" +
            "ALOAD 3\n" +
            "ALOAD 4\n" +
            "INVOKESPECIAL resources/OnMethodTest.$btrace$native$nativeWithoutReturn (ILjava/lang/String;[J[Ljava/lang/Object;)V\n" +
            "RETURN\n" +
            "MAXSTACK = 5\n" +
            "MAXLOCALS = 5\n" +
            "\n" +
            "public native $btrace$native$nativeWithoutReturn(ILjava/lang/String;[J[Ljava/lang/Object;)V\n" +
            "\n" +
            "// access flags 0xA\n" +
            "private static $btrace$traces$onmethod$NativeWithoutReturn$nMethod(Ljava/lang/Object;)V\n" +
            "@Lcom/sun/btrace/annotations/OnMethod;(clazz=\"/.*\\\\.OnMethodTest/\", method=\"nativeWithoutReturn\")\n" +
            "@Lcom/sun/btrace/annotations/Self;() // parameter 0\n" +
            "TRYCATCHBLOCK L0 L1 L1 java/lang/Throwable\n" +
            "GETSTATIC traces/onmethod/NativeWithoutReturn.runtime : Lcom/sun/btrace/BTraceRuntime;\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.enter (Lcom/sun/btrace/BTraceRuntime;)Z\n" +
            "IFNE L0\n" +
            "RETURN\n" +
            "L0\n" +
            "LDC \"args\"\n" +
            "INVOKESTATIC com/sun/btrace/BTraceUtils.println (Ljava/lang/Object;)V\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "RETURN\n" +
            "L1\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.handleException (Ljava/lang/Throwable;)V\n" +
            "INVOKESTATIC com/sun/btrace/BTraceRuntime.leave ()V\n" +
            "RETURN\n" +
            "MAXSTACK = 1\n" +
            "MAXLOCALS = 1"
        );
    }
}